# er2 interface
#
# The er2 class here is the base class for the
#   real, simulated, client/server, and other er2 options
#

import time              # standard
import math              # standard
import random            # standard
import sys               # standard
import sensorSimulator as sim   # our world simulator

GLOBAL_DELTA = 19.0      # 1/2 the distance between the wheels in cm

#
# A class to hold the (x, y, thr) pose of the robot
#
class ER2OdometricPosition:

    def __init__(self, x_in=0., y_in=0., thr_in=0.):
        self.x = x_in
        self.y = y_in
        self.thr = thr_in # radians from 0 to 2pi - change to thr sometime...
	global GLOBAL_DELTA
        self.DELTA = GLOBAL_DELTA

    def pr(self):
        print 'x:', self.x
        print 'y:', self.y
        print 'th (degrees):', self.thr*180.0/math.pi
        print 'th (radians):', self.thr

    def set(self,x_in,y_in,thr_in):
        self.x = x_in
        self.y = y_in
        self.thr = thr_in

    def reset(self):
        self.set(0.,0.,0.)

    #
    # here are the er1's kinematics!
    #
    def diffDriveKinUpdate(self, left_cm, right_cm):
        """This method takes in
               left_cm  = the number of cm the left wheel has turned
               right_cm = the number of cm the right wheel has turned
                          (forward is positive, reverse is negative)

            and it changes the calling object to reflect this motion
        """
        #
        # debugging
        #
        #print 'in Kin, l,r are', left_cm, right_cm

        # let's leave if there is nothing to do...
        if left_cm == 0 and right_cm == 0:
            return

        
        # compute the radius of curvature (ROC)
        #
        # this assumes that the robot has been traveling with constant
        # radius of curvature, which should be mostly satisfied
        # as long as we update the position every time that the velocity
        # is changed
        #
        # ROC = DELTA * (R + L) / (R - L), where DELTA is 1/2 distance between the wheels
        ROCdenominator = right_cm - left_cm
        ROCnumerator = right_cm + left_cm

        # if the denominator's value is too small relative to the numerator, the robot is 
        # essentially traveling in a straight line - the following ratio of 10000:1 means
        # that we will consider anything within 0.005 of a degree a straight line - 
        # likely undetectable, given all of the other sources of odometric error

        if ROCdenominator == 0 or math.fabs(ROCnumerator/float(ROCdenominator)) > 10000.0:
            # theta does not change
            self.x += math.cos(self.thr)*left_cm # left_cm ~= right_cm
            self.y += math.sin(self.thr)*left_cm

        else:  # left_cm ~= right_cm
            #global gvel_left_cmpersec, gvel_right_cmpersec, gtime
            #print 'gvel:l,r,time are', gvel_left_cmpersec, gvel_right_cmpersec, gtime
	    #print 'gpos before is', self.x, self.y, self.theta*180.0/math.pi
	    #print 'right_cm, left_cm are', right_cm, left_cm
            #print 'ROCden, ROCnum are', ROCdenominator, ROCnumerator
	    #
            # otherwise, we have to figure out the ROC and the angle turned
            # the angle, theta, is computed based on the wheel that turned further
            #
            # that wheel rotated around a radius of (DELTA + fabs(ROC))
            #
            # theta = max(fabs(L),fabs(R)) / (DELTA + fabs(ROC)) 
            #
            # DELTA has to be in the same units as w1 and w0...
            ROC = self.DELTA * float(ROCnumerator) / float(ROCdenominator)
            #print 'ROC is', ROC
            d_thr = max(math.fabs(right_cm),math.fabs(left_cm)) / (self.DELTA + math.fabs(ROC))
            # but theta's sign is not necessarily correct...
            #
            # we will use a positive value to indicate counterclockwise rotation (left)
            # with respect to the robot's "forward-facing" pose
            #
            # the opposite (negative) rotation occurs when the left wheel has
            # traveled further
            #
            if left_cm > right_cm:
                d_thr = -d_thr

            # new conversion
            #print 'theta is', theta*180.0/math.pi;

            # first, we compute the canonical distance traveled
            # assume the robot is facing the positive x direction
            # and that positive y is to the left and the robot
            # is at (0,0)
            #
            # hooray for ASCII art!
            #
            #                ^ +x (aligned with self.theta)
            #                |
            # L=left_cm  ^   |   ^  right_cm=R
            #            |   |   |
            #     +y <---W1--+--W0--- -y (perp to self.theta)
            #                |
            #                |       DELTA = 1/2 distance from W1 to W0
            #                | -x
            #
            # The wheels W1 and W0 give rise to positive differences
            # when they are headed forward along the positive x axis
            #
            # then the center of the circle along which it's traveling
            # is at (0,ROC)
            # 
            # and the resulting position after the arc is
            # (0,ROC) + (ROC*sin(theta),-ROC*cos(theta))
            #
	    # kinematics error #2: used absolute value below...
            deltax = 0.0 + ROC*math.sin(d_thr)
            deltay = ROC - ROC*math.cos(d_thr)
            #print 'local deltax and deltay are', deltax, deltay

            # now, we need to convert this local-coordinate deltax and deltay
            # into the global frame - 
            # deltax is the change along self.theta
            # deltay is the change perpendicular to self.theta, along self.theta + pi/2
            #
            x_from_deltax = math.cos(self.thr) * deltax
            # are you joking me!! the value on the right below was deltay - ouch!
            y_from_deltax = math.sin(self.thr) * deltax
            x_from_deltay = math.cos(self.thr + math.pi/2.0) * deltay
            y_from_deltay = math.sin(self.thr + math.pi/2.0) * deltay

            #print 'global deltax and deltay are', (x_from_deltax + x_from_deltay), (y_from_deltax + y_from_deltay)

            # adjust position
            self.x += (x_from_deltax + x_from_deltay)
            self.y += (y_from_deltax + y_from_deltay)
            self.thr += d_thr

            # make sure we're less or equal to 2pi
            while self.thr > math.pi*2.:
                self.thr -= math.pi*2.
            # make sure we're greater than or equal to 0
            while self.thr < 0:
                self.thr += math.pi*2.

            #print for debugging...
            #print 'l,r,ROC,thd are', left_cm, right_cm, ROC, self.theta*180.0/3.14159
	    #print 'gpos after is', self.x, self.y, self.theta*180.0/math.pi, '\n'

        #print 'current pos is'
        #self.pr();

#
# the base class (struct, really) for er2 sensor frames: one snapshot in time
#
class ER2sensorframe:

    def __init__(self, pose, velocities, otherdata={}):
        self.pose = pose
        self.velocities = velocities
        self.timestamp = time.clock()
        self.otherdata = otherdata

    def getData(self, style)
        """ this method will return data depending on the "style"
            if style == 0:
              ( x, y, thr, otherdata['sens'], timestamp, otherdata['realp'] )
              for legacy reasons...
            if style == 1:
        """
        if style == 0:
            return ( self.pose.x, self.pose.y, self.pose.thr, \
                     self.otherdata['sens'], self.timestamp,  \
                     self.otherdata['realp'] )
        else:
            return 'unsupported style in ER2sensorframe.getData'


#
# the base class for er2 interfaces...
#
class er2:

    def __init__(self, mapfile=None):
        global GLOBAL_DELTA  
        self.DELTA = GLOBAL_DELTA  # 1/2 the distance bewteen the wheels = ~19cm
        # DELTA is needed to convert from deg per sec to wheel velocities...

        # here is the speed limit of each wheel in cm/sec
        self.SPEEDLIMIT = 20    # 20 for real robot, 2000 for simulated...
        
        # the two wheel velocities
        self.gvel_left_cmpersec = 0.0
        self.gvel_right_cmpersec = 0.0
        
        # to hold the robot's estimated (odometry-based) position
        self.gpos = ER2OdometricPosition()

        # pose and sensor history with timestamps
        # this should be a list of ER2sensorframes
	self.fullHistory = []

        # get the map - must be done before the call to self.InitializeRobot,
        # which calls update, which needs the map...
        self.map = sim.Map( mapfile )


    def updateHistory(self, otherdata={}):
        """ updateHistory takes a current snapshot of the robot's
            internal and sensor values and stores them in an ER2sensorframe
            and then adds that object to the self.fullHistory list 
        """
        self.fullHistory.append( ER2sensorframe( self.gpos, \
                                                 (self.gvel_left_cmpersec, \
                                                  self.gvel_right_cmpersec), \
                                                 self.otherdata, \
                                               ) )

    def clearHistory():
        """ set it to the single ER2sensorframe currently at the end """
        if len(self.poseHistory) > 0:
            self.poseHistory = self.poseHistory[-1:]
        return None
    
    def convertCmDegToWheelVelocities(self, cmpersec, degpersec):
        """ set robot velocities in cmpersec and degpersec

            this function implements the kinematics...
        """
        # need to invert the kinematics to get the wheel rotations in cm per second
        # the translational component is easy
        left_trans_cm_sec = cmpersec
        right_trans_cm_sec = cmpersec
    
        # the rotational component is not too bad 
        # note that if the translational component is 0, this
        # allows for a maximum of about 60 degrees per second...
        # with our safety maximum of 20 cm/second per wheel
    
        # ok, I think this does not combine rotation and translation
        # in the most natural way
        left_rot_cm_sec = - self.DELTA * (degpersec * math.pi / 180.0 )
        right_rot_cm_sec = self.DELTA * (degpersec * math.pi / 180.0 )
        # print 'cm, deg', cmpersec, ',', degpersec
        # print 'left,right trans', left_trans_cm_sec, ',', right_trans_cm_sec
    
        # add them up
        left_cm_sec = left_trans_cm_sec + left_rot_cm_sec
        right_cm_sec = right_trans_cm_sec + right_rot_cm_sec
        # print 'left,right total', left_cm_sec, ',', right_cm_sec
    
        # send off to robot
        return (left_cm_sec, right_cm_sec)
    
    def SetWheelVelocities(self, vels)
        """ set robot wheel velocities: (left_cm_sec, right_cm_sec)

            this checks the robot's own notion of its wheel
            velocities and does nothing if they are identical
        """
        # unpack the velocities
        left_cm_sec, right_cm_sec = vels

        # why change if there's no change to make?
        if self.gvel_right_cmpersec == right_cm_sec and \
           self.gvel_left_cmpersec == left_cm_sec: return

        # set the velocities and wait for next chance to update the odometry
        self.gvel_right_cmpersec = right_cm_sec
        self.gvel_left_cmpersec = left_cm_sec


    def speedLimitWheelVelocities( self, wheelVelocities, speedlimit ):
        """ return a speed-limited (scaled) pair of velocities
            given the desired speed limit 
        """
        # unpack velocities
        left_cm_sec, right_cm_sec = wheelVelocities

        maxSpeed = max(left_cm_sec, right_cm_sec, -left_cm_sec, -right_cm_sec)
        speedlimit = max(0, speedlimit)
    
        if maxSpeed > speedlimit
            left_cm_sec *= speedlimit/maxSpeed
            right_cm_sec *= speedlimit/maxSpeed

        return (left_cm_sec, right_cm_sec)
    
    
    def Go(self, cmpersec, degpersec=0.0):
        """ set robot velocities in cmpersec and degpersec """
        wheelVelocities = self.convertCmDegToWheelVelocities(cmpersec,degpersec) 
        limitedVelocities = self.speedLimitWheelVelocities( wheelVelocities, self.SPEEDLIMIT )
        self.SetWheelVelocities( limitedVelocities )
        return None

    def SetOdometry(self, x, y, thr):
        """ we tell the robot where it is! """
        self.gpos.set(x,y,thr)
        self.updateHistory()
        return None
    
    def GetOdometry(self):
        """ returns (x, y, thr) """
        return [self.gpos.x, self.gpos.y, self.gpos.thr]

    def GetAllDataForP( self, xp, yp, thrp ):
        """ returns a 2-tuple whose first element is
            the sensor data for a robot (or particle)
            at (xp, yp, thrp) based on the robot's map

            the second element is the number of wall-hits
            of the segment from the input pose to the
            point (-424242.42, -424242.42), presumably off the map!
        """
        numhits = self.map.numHits(xp,yp,thrp)
        R = self.map.findER1ranges(xp,yp,thrp)
        return (R,numhits)

    def GetAllData(self, style=0):
        """ returns a tuple of data from the sensors and robot """
        self.updateHistory()
        return self.fullHistory[-1].getData(style)
    
    def PowerOff(self):
        """ sets velocities to zero """
        self.SetWheelVelocities(0.0,0.0)
        return None
    
    def CloseRobot(self):
        """ closes things down """
        self.SetWheelVelocities(0.0,0.0)
        return None


#
# for a simulated robot
#
class er2sim(er2):

    def __init__(self, mapfile):
        er2.__init__(mapfile)
        self.SPEEDLIMIT = 2000.0   # faster than 20 for the simulated robot...
        
    def SetWheelVelocities( limitedVelocities )
    def SetOdometry(self, x, y, thr):
    def GetOdometry(self):
    def GetAllData(self, style=0):
    def PowerOff(self):
    def CloseRobot(self):


class er2real(er2):

class er2realabs(er2real)
        # need ABS stuff here!! ??

class er2client(er2):

class er2thread(er2):

