# ER2 server
import socket
import time
import sys
import er2sim as er2
import abs                     # abs = acroname brain stem

# set up networking parameters
HOST = '127.0.0.1'             
HOST = '169.254.87.131'        
PORT = 5050              

# s will be the name of our connection
s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
s.bind((HOST, PORT))
s.settimeout(0.1)            #s.setblocking(0)
s.listen(1)

# variables for tracking the status of the server connection
accepted = False
count = 0
conn = None

# create an ER2 object
robot =  er2.ER2(PORT=0, mapfile="mapSquare.txt")
# print robot.GetOdometry()

# initialize the abs interface
COM_PORT_NUMBER = 1
#### abs.initializeBoard( COM_PORT_NUMBER )
abs = None


def handleData( data ):
    """ this function, handleData, 
          1) translates the message, data into robot/sensor actions
          2) formats the results (if any) and returns the formatted string
    """
    # print 'input data is ', data
    # make sure we have a string...
    if type(data) == type('s'):

        # parse data
        listdata = data.strip().split('_')  # case sensitive!
        # print 'listdata is', listdata

        if len(listdata) > 0:

            message = listdata[0]    # the name of the request

            # currently we handle the messages
            #
            # GetOdometry
            # SetOdometry
            # GetAllData
            # Go
            #
            # others that should be handled, but not yet
            #
            # setPTvel
            # setPTpos
            # getPTpos
            #
            if message == 'GetOdometry':
                o = robot.GetOdometry()
                reply = '%.3f_%.3f_%.3f' % (o[0], o[1], o[2])
                #print 'o is', o
                #print 'reply is', reply
                return reply

            elif message == 'GetAllData':
                # get all the data from the abs
                # return it...
                # for next demo, use 4 IR range sensors
                # convert to cm before this point
                r1 = 10
                r2 = 12
                r3 = 20
                r4 = 42
                reply = '%.2f_%.2f_%.2f_%.2f' % (r1, r2, r3, r4)
                return reply

            # note that GetAllDataFrom... can always be done locally
            # so it does not need to be implemented here...

            elif message == 'SetOdometry':
                # the format will be 'SetOdometry_12.42_-42.42_1.5742'
                # in units of cm, cm, and radians
                if len(listdata) != 4:
                    print 'listdata of', listdata, 'has wrong length in SetOdometry'
                    print 'Not setting the robot\'s odometry'
                    reply = 'no_reply'
                    return reply

                # here, it must have been formatted correctly
                x = float(listdata[1])
                y = float(listdata[2])
                thr = float(listdata[3])
                robot.SetOdometry( x, y, thr )

                reply = 'no_reply'
                return reply


            elif message == 'Go'
                # the format will be Go_10.42_2.42
                # the first will be in cm/sec
                # the second will be in deg/sec
                if len(listdata) != 3:
                    print 'listdata of', listdata, 'has wrong length in Go'
                    print 'Not changing robot\'s speed'
                    reply = 'no_reply'
                    return reply
                
                # here, it must have been formatted correctly
                cmpersec = float(listdata[1])
                degpersec = float(listdata[2])
                robot.Go(cmpersec, degpersec)

 
            else:
                print 'message', message, 'not understood in handleData'
                print 'returning \'bad_message\''
                return 'bad_message'


        else:
            return data
    else:
        print 'Non string data', data, 'seen in handleData'
        print 'returning \'bad_data\''
        return 'bad_data'


try:   # so that we catch keyboard exceptions and close objects safely

    # listen for connections...
    while True:
    
        # reset ER2 and ABS objects to safe, steady states
        robot.Go( 0.0, 0.0 ) # just to test
        # abs - reset pan/tilt head (if there is one)
    
        print 'Listening for connections ...'
    
        # the listening loop
        while True:

            try:
                conn, addr = s.accept()
                print 'Connected to', addr    # , 'at count', count
                break
            except socket.error:
                count += 1
                time.sleep(0.1)

        # now, we begin the request-response loop
        robot.Go( 10.0, 0.0 ) # just to test
        
        while True:
            try:
                data = conn.recv(1024)
                if len(data) < 1: break   # breaks if the connection is broken
                if data[0] == 'q': break  # a controlled quit
    
                # here is where the data is handled
                reply = handleData( data )
                conn.send(reply)   # send back the desired reply
    
            except:
                # no data received... 
                time.sleep(0.01)
    
        conn.close()   # close the broken connection and listen again

except KeyboardInterrupt:

    print 'Keyboard Interrupt... closing robot and sensors'

    # close robot objects gracefully and exit
    robot.CloseRobot()
    # abs
    # abs.close()

if conn != None:  # and close the wireless connection, too
    conn.close()

