# Hello Mac -
#
# I apologize for not getting you this update of
# the mcc code sooner. I have experimented with
# a general function solver (in addition to our
# linear version): each works in some cases, but
# not others, though it's not really possible,
# as far as I can tell, to pin down when the
# general function solve fails.

# However, to run everything you'll need the
# scientific python modules - I downloaded
# the Python 2.3 "Enthought Edition" (you'll
# find it if you google for that).

# From this directory, run
# cmd prompt> start s.bat
# to start up the camera_image, robot_image, and
# world-view windows

# Then run
# cmd prompt> python mcc.py
# and you'll have to zoom out (right-click-drag)
# to see the square world and the robot inside it

# you can left-click to drag the robot around
# type 'R' to show the robot's "real" view
# hold the 'r' (rotate) key and drag the robot to
#   change its rotation

# you can create a camera (one of the mcc particles)
# by holding the 'c' key down and left-clicking-and-dragging
# to determine its center and direction

# when the robot and camera are where you want them,
# type 'P' in the main window to update the pose -
# a green camera (t-shape) will appear where the
# program thinks the robot is - sometimes it's right,
# and sometimes it's not.

# Let's think more about where to go with this...
# At the moment, at least, I feel confident about creating
# images for a paper. Still not done are
#   - guessing correspondences
#   - trials with noise
#   - using field-of-view information

# Talk to you later,

# Zach

# mcc.py
#
# shift-left-drag to zoom in and out of the visualizer

# python modules with things we need...
import math     # all the familiar functions
import er2sim as er2       # the robot model
import visClient           # to send messages to the Visualizer window

import poseEstimator       # the scipy depedencies

# out map file name
mapfilename= "bugMap.txt"
mapfilename= "mapColors.txt"

# our robot
robot = er2.ER2(PORT=0, mapfile=mapfilename)

# our connection to the robot visualizer
vc = visClient.VisualizerClient()   # using default host and port
vc.exchMessage('loadmap_%s' % mapfilename)

# our connection to the image display
vc_image = visClient.VisualizerClient(PORT=5009)
cam_image = visClient.VisualizerClient(PORT=5010)

# the starting robot velocities: translational (cm) and rotational (deg)
v_cm_old = 0.0
v_deg_old = 0.0

# these are constants for accessing the sense range values
BaRi = 0    # the index of the Back Right sensor reading
BaLe = 1    # the index of the Back Left sensor reading
FrLeSi = 2  # the index of the Front Left side-looking sensor reading
FrLe = 3    # the index of the Front Left front-looking sensor reading
Fr   = 4    # the index of the Front sensor reading
FrRi = 5    # the index of the Front Right front-looking sensor reading
FrRiSi = 6  # the index of the Front Right side-looking sensor reading

def frontThree(R):
    return R[3:6]

def angleCompare( thd1, thd2, window ):
    # make sure the angles are between 0.0 and 360.0
    while thd1 > 360.0: thd1 -= 360.0
    while thd2 > 360.0: thd2 -= 360.0
    while thd1 <   0.0: thd1 += 360.0
    while thd2 <   0.0: thd2 += 360.0
    # figure out their relative positions on the unit circle
    smaller = min(thd1,thd2)
    larger  = max(thd1,thd2)
    # figure out the two possible differences
    oneway = larger - smaller
    otherway = (360.0-larger) + smaller
    # is the difference within the window?
    return min(oneway,otherway) <= window

# get images
def getImage(x,y,thr,robot):
    thd = 180.0 * thr / math.pi
    exactim = robot.map.getExactImage(x,y,thd) 
    #print 'exact im at', x, y, thd
    #print 'is '
    imstr = "i"
    for seg in exactim:
        x1 = seg.x1 
        x2 = seg.x2 
        if x1 > x2: x1, x2 = x2, x1 
        #print '   ', x1, x2, seg.r, seg.g, seg.b
        imstr = imstr + "_" + str(x1) + "_" + str(x2) +\
                        "_" + str(seg.r) + "_" + str(seg.g) + "_" + str(seg.b)
        #print 'imstr is', imstr
    return imstr

# state constants
GoForward = 0
TurnLeft90 = 1
TurnRight90 = 2
Stop = 8
Goal = 9
MAX_STATE = 10

# initial state
state = GoForward
state = Stop

# constants
FRONT_THRESH = 20

# velocities for each state
velocities = [0]*MAX_STATE
velocities[GoForward] = (42.0, 0.0) # cm/sec, deg/sec
velocities[TurnLeft90] = (0.0, 20.0)
velocities[TurnRight90] = (0.0, -20.0)
velocities[Stop] = (0.0, 0.0)
velocities[Goal] = (0.0, 0.0)

# here we're using x = 105.0, y= 39.0, thr = -0.7854
x = 105.0
y = 39.0
thr = -0.7854
robot.SetOdometry(x,y,thr)
imstr = getImage(2.02,-4.8, math.radians(-67.27),robot)
print 'imstr is', imstr
cam_image.exchMessage(imstr)

# main robot-control loop
while True:

    # S = sense: what do my sensors tell me?
    (x,y,thr,(R,stalled),timestamp,realPose) = robot.GetAllData()
    thd = thr*180.0/math.pi
    #print 'x,y,thd are', x, y, thd
    #print 'stalled, timestamp, R are', stalled, timestamp, R


    # P = plan

    # figure out what state to be in
    # print a message if we are changing state
    if state == GoForward:
	if min(frontThree(R)) < FRONT_THRESH:
            state = TurnRight90
            print 'I am changing to state TurnRight90'
            desiredAngled = thd - 93
            angleWindowd = 3

    elif state == TurnRight90:
            if angleCompare( thd, desiredAngled, angleWindowd ):
                state = GoForward
                print 'I am changing to state GoForward'

    else:
        # I'm lost - just stop
        state = Stop

    # set the velocities appropriate to the state we are in
    v_cmpersec =  velocities[state][0]
    v_degpersec = velocities[state][1]


    # A = act: send velocities to the robot only if they've changed
    if v_cmpersec != v_cm_old or v_degpersec != v_deg_old:
        robot.Go(v_cmpersec, v_degpersec)
        v_cm_old = v_cmpersec
        v_deg_old = v_degpersec


    # this is a problem, because the various exchanges
    # miss potential incoming messages!

    # visualize: prepare data for the graphics window (robot-depedent)
    #   there is probably a cleverer python way to do this...

    # turn off orange range sensor bars for now...
    R = [0,0,0,0,0,0,0]
    msg = 'rM_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f' %    \
          (x,y,thr,R[0],R[1],R[2],R[3],R[4],R[5],R[6])
    retmsg = vc.exchMessage(msg)
    if retmsg != 'hi 42': 
        print 'retmsg is', retmsg
        msglist = retmsg.split('_')
        if msglist[0] == 'pose':
            print 'pose!'
            newx = float(msglist[1])
            newy = float(msglist[2])
            newthr = float(msglist[3])
            robot.SetOdometry(newx, newy, newthr)
	elif msglist[0] == 'cpose':
            print 'cpose!'
            camx = float(msglist[1])
            camy = float(msglist[2])
            camthr = float(msglist[3])
            # show camera's image for fun
            imstr = getImage(camx, camy, camthr, robot)
            cam_image.exchMessage(imstr)
	elif msglist[0] == 'showRobotImage':
            imstr = getImage(x,y,thr,robot)
            print 'imstr is', imstr
            vc_image.exchMessage(imstr)
        elif msglist[0] == 'poseAdjust':
            # get the camera's location from the message 
            if len(msglist) == 4:
                camx = float(msglist[1])
                camy = float(msglist[2])
                camthr = float(msglist[3])
            else:
                # we'll use the origin
                camx = 0.0
                camy = 0.0
                camthr = 0.0

            # display the camera's image for fun
            imstr = getImage(camx, camy, camthr, robot)
            cam_image.exchMessage(imstr)

	    # where is the robot?
	    #print
	    #print 'The robot\'s pose:', x, y, thd
	    
	    # for good measure, we'll show the robot's "real" image
            imstr = getImage(x,y,thr,robot)
            vc_image.exchMessage(imstr)

            # get the real image and real image
            # features (this is the robot's)
            feats = robot.map.getExactImageFeatures(x,y,thd) 
            #print 'feats is', feats

            # separate into image features and world features
            imf = []
            WF = []
            for k in feats.iterkeys():
                # impose field-of-view constraints
                # the sensor simulator provides FOV = 1000, I think !
                # FOV = 1.0
                #FOV = 1000.0
                #if -FOV <= k <= FOV:
                    imf.append(k)
                    WF.append(feats[k])
            print 'imf is', imf
            print 'WF is', WF

            # green correctWF = [(200,-200),(200,0),(0,-200)]
            # turn clockwise ltblue WF = [(0,-200),(200,-200),(-200,-200)]

            # we create our own correspondences here!!
            WF = [(0,-200),(200,-200),(-200,-200)]

            # then, add noise
            #     none, for now
            imf_noisy = imf[:]

            # then, decide which image features correspond
            # to which world features
            #     we use known correspondences, for now


            # then, use the camera's pose as the starting
            # point for the solution
            SP =  [camx, camy, camthr]
            print 'SP is', SP

            # find the corrected pose
            #cx, cy, cthr = poseEstimator.poseEstimate( WF, imf_noisy, SP )
            #cx, cy, cthr = poseEstimator.poseEstimate12( WF, imf_noisy, SP )
            cx, cy, cthr = poseEstimator.poseEstimate3( WF, imf_noisy, SP )
            cthd = cthr*180.0/math.pi
            print 'result is', cx, cy, cthd
            camera = "c_%f_%f_%f_%f_%f_%s" % (cx,cy,cthd,10,8,"green")
            vc.exchMessage(camera)



