# mclGood.py
#

START_ON_ROBOT = True
USE_OUTOFBOUNDS = True

USE_RANGE_VALUES = False
USE_ONE_RANGE_VALUE = False
USE_ONE_RANGE_VALUE_NG = True

NOISE_PERCENTAGE = 0.50   # 0.25 means 25% not 0.25% !
ANOISE_PERCENTAGE = 0.10   # 0.25 means 25% not 0.25% !

# python modules with things we need...
import time                # fun with clocks!
import math                # all the familiar functions
import random              # maybe we'll use it, maybe we won't
import er2sim as er2       # the robot model
import visClient           # to send messages to the Visualizer window

# out map file name
mapfilename= "bugMap.txt"
mapfilename= "mapColors.txt"

# our robot
robot = er2.ER2(PORT=0, mapfile=mapfilename)

# our connection to the robot visualizer
vc = visClient.VisualizerClient()   # using default host and port
vc.exchMessage('loadmap_%s' % mapfilename)

# our connection to the image display
#vc_image = visClient.VisualizerClient(PORT=5009)
#cam_image = visClient.VisualizerClient(PORT=5010)

# the starting robot velocities: translational (cm) and rotational (deg)
v_cm_old = 0.0
v_deg_old = 0.0

# these are constants for accessing the sense range values
BaRi = 0    # the index of the Back Right sensor reading
BaLe = 1    # the index of the Back Left sensor reading
FrLeSi = 2  # the index of the Front Left side-looking sensor reading
FrLe = 3    # the index of the Front Left front-looking sensor reading
Fr   = 4    # the index of the Front sensor reading
FrRi = 5    # the index of the Front Right front-looking sensor reading
FrRiSi = 6  # the index of the Front Right side-looking sensor reading

def frontThree(R):
    return R[3:6]

def angleCompare( thd1, thd2, window ):
    # make sure the angles are between 0.0 and 360.0
    while thd1 > 360.0: thd1 -= 360.0
    while thd2 > 360.0: thd2 -= 360.0
    while thd1 <   0.0: thd1 += 360.0
    while thd2 <   0.0: thd2 += 360.0
    # figure out their relative positions on the unit circle
    smaller = min(thd1,thd2)
    larger  = max(thd1,thd2)
    # figure out the two possible differences
    oneway = larger - smaller
    otherway = (360.0-larger) + smaller
    # is the difference within the window?
    return min(oneway,otherway) <= window

# state constants
GoForward = 0
TurnLeft90 = 1
TurnRight90 = 2
Stop = 8
Goal = 9
MAX_STATE = 10

# initial state
state = GoForward
print 'I am starting in state GoForward'

# constants
FRONT_THRESH = 20

# velocities for each state
velocities = [0]*MAX_STATE
velocities[GoForward] = (42.0, 0.0) # cm/sec, deg/sec
velocities[TurnLeft90] = (0.0, 20.0)
velocities[TurnRight90] = (0.0, -20.0)
velocities[Stop] = (0.0, 0.0)
velocities[Goal] = (0.0, 0.0)


# PL is the particle list
# each particle is a list [x,y,thr,p], where 
# p represents a relative probability (need not sum to 1)
def randomParticleStart(x=0.0, y=0.0, thr=0.0):
    NUM_PARTICLES = 100
    PL = []
    if START_ON_ROBOT:
        xrange = 0
        yrange = 0
        thrrange = 0
    else:
        xrange = 200
        yrange = 200
        thrrange = 0
        x = 0.0
        y = 0.0
        thr = 0.0

    for i in range(0,NUM_PARTICLES):
        PL.append( [ random.uniform(x-xrange,x+xrange), \
                     random.uniform(y-yrange,y+yrange), \
                     random.uniform(thr-thrrange,thr+thrrange), \
                     1.0 ] )
    return PL


def computeSmallestAngleDifference( fromThr, toThr ):
    # make sure the angles are between 0.0 and 360.0
    twopi = 2*math.pi
    while fromThr > twopi: fromThr -= twopi
    while toThr > twopi: toThr -= twopi
    while fromThr <  0.0: fromThr += twopi
    while toThr <  0.0: toThr += twopi
    # figure out their relative positions on the unit circle
    smaller = min(fromThr, toThr)
    larger  = max(fromThr, toThr)
    # figure out the two possible differences
    oneway = larger - smaller
    otherway = (twopi-larger) + smaller
    # is the difference within the window?
    return min(oneway,otherway)

def computeSmallestSignedAngleDifference( fromThr, toThr ):
    posDiffr = computeSmallestAngleDifference( fromThr, toThr )
    # but we need to sign this to go from fromThr to toThr
    add = computeSmallestAngleDifference( fromThr + posDiffr, toThr )
    sub = computeSmallestAngleDifference( fromThr - posDiffr, toThr )
    if add < sub:
        return posDiffr
    else:
        return -posDiffr

def computeTotalRobotMotion( oldLoc, newLoc ):
    oldx, oldy, oldthr = oldLoc
    x, y, thr = newLoc
    # we return thr1, dist, thr2
    # where
    # thr1 is the smallest angle in radians from oldThr
    #         to the direction of the vector (x,y)-(oldx,oldy)
    # dist is the distance between (x,y) and (oldx, oldy)
    # thr2 is the smallest angle in radians from the direction
    #         of the vector mentioned above to thr
    motion_thr = math.atan2(y-oldy, x-oldx)
    thr1 = computeSmallestSignedAngleDifference( oldthr, motion_thr )
    dist = math.sqrt( (x-oldx)**2 + (y-oldy)**2 )
    thr2 = computeSmallestSignedAngleDifference( motion_thr, thr )
    return (thr1, dist, thr2)

def addNoise( thr1, dist, thr2 ):
    # return a triple: the same nominal values, with noise
    #
    NOISE_TYPE = 'unif_perc'
    #
    if NOISE_TYPE == 'none':
        return (thr1,dist,thr2)
    elif NOISE_TYPE == 'unif_perc':  # uniform noise as a percent of value
	low = min(-ANOISE_PERCENTAGE*thr1, +ANOISE_PERCENTAGE*thr1)
	high = max(-ANOISE_PERCENTAGE*thr1, +ANOISE_PERCENTAGE*thr1)
        thr1n = random.uniform(low,high)
	low = min(-NOISE_PERCENTAGE*dist, +NOISE_PERCENTAGE*dist)
	high = max(-NOISE_PERCENTAGE*dist, +NOISE_PERCENTAGE*dist)
        distn = random.uniform(low,high)
	low = min(-ANOISE_PERCENTAGE*thr2, +ANOISE_PERCENTAGE*thr2)
	high = max(-ANOISE_PERCENTAGE*thr2, +ANOISE_PERCENTAGE*thr2)
        thr2n = random.uniform(low,high)
        return (thr1 + thr1n,dist + distn,thr2 + thr2n)
    elif NOISE_TYPE == 'uniform':  # real uniform noise - not much sense here
        pass
    elif NOISE_TYPE == 'tri_perc':  # real uniform noise - not much sense here
        pass
    elif NOISE_TYPE == 'gauss_perc':  # real uniform noise - not much sense here
        pass

def moveParticles( PL, thr1, dist, thr2 ):
    for p in PL:
        # add noise to the values... to find thr1n, distn, thr2n
        thr1n, distn, thr2n = addNoise( thr1, dist, thr2 )        
        p[0] += distn*math.cos(p[2] + thr1n)
        p[1] += distn*math.sin(p[2] + thr1n)
        p[2] += (thr1n + thr2n)  # add the angles
    return PL

def resampleParticles( PL ):
    # what was the clever way to do this again?
    # we want to get l particles out, right...
    newPL = []
    # need the sum of the "probabilities"
    sum = 0.0
    for part in PL: sum += part[3]
    # that's 100%, now start at 0.5
    deltaThresh = sum/len(PL) # amount to raise curThresh by each time
    curThresh = deltaThresh/2.0 # seems like a safe starting point
    curPartSum = 0.0
    curPartIndex = 0
    # two nested while loops
    while curPartIndex < len(PL):  # particle loop
        curPartSum += PL[curPartIndex][3]  # add next particle's weight
        while curPartSum > curThresh:  # sampling/probability loop 
            newPL.append( PL[curPartIndex][:] ) # new copy
	    newPL[-1][3] =  1.0
            curThresh += deltaThresh  # invariant: append1 for each! curThresh
        curPartIndex += 1  # next particle
    return newPL

def rangeWeight( r1, r2 ):
    # returns the update weight for these two ranges
    # if equal == 1
    # if outside 100% error == 0.1
    MIN_WEIGHT = 0.1
    # linear in between
    rmin = min( r1, r2 )
    rmax = max( r1, r2 )
    if rmin <= 0.0 or rmax > 2.0*rmin: return MIN_WEIGHT
    # percError should be between 0.0 and 1.0
    percError = (float(rmax)/float(rmin)) - 1.0
    # map to the range 1.0 -> 0.1
    relweight = (1.0-percError) * 1.0 + percError * 0.1
    # now, if within 5 cm, 1.0
    if rmax-rmin < 5.0: return 1.0
    # if outside 50 cm, 0.1*relweight
    absweight = 1.0
    if rmax-rmin > 50.0:
        absweight = 0.1/(rmax-rmin-49.0)
    return absweight*relweight

def updateParticles( PL, robot, ALL_SENSORS ):
    (x,y,thr,(R,stalled),timestamp,(rx,ry,rthr)) = ALL_SENSORS
    for part in PL:
        # we need to update its probability
        # first, we get its sensor state
        RP, numhitsP = robot.GetAllDataFrom(part[0],part[1],part[2]) 
        #
        # compare stalled and numhits (stalled is False!)
        if USE_OUTOFBOUNDS:
            if numhitsP % 2 == 0:
                part[3] *= 0.0001
        #
        # compare R and RP
        # first, just forward...

        if USE_ONE_RANGE_VALUE_NG and R[Fr] - RP[Fr] > 10.0:
            #for i in range(0,7):
                part[3] *= 0.001

        if USE_ONE_RANGE_VALUE:
            #for i in range(0,7):
                part[3] *= rangeWeight( RP[Fr], R[Fr] )

        if USE_RANGE_VALUES:
            for i in range(0,7):
                part[3] *= rangeWeight( RP[i], R[i] )
        #BaRi = 0    # the index of the Back Right sensor reading
        #BaLe = 1    # the index of the Back Left sensor reading
        #FrLeSi = 2  # the index of the Front Left side-looking sensor reading
        #FrLe = 3    # the index of the Front Left front-looking sensor reading
        #Fr   = 4    # the index of the Front sensor reading
        #FrRi = 5    # the index of the Front Right front-looking sensor reading
        #FrRiSi = 6  # the index of the Front Right side-looking sensor reading
    # now resample the particles
    PL = resampleParticles( PL )
    return PL

# update the particles here...
PL = randomParticleStart()
# these lines will send the message to redraw them...
msg = vc.createParticleMessage(PL)
vc.exchMessage(msg)
# remember where the robot was when this reset occurred
lastUpdateLoc = (0.0,0.0,0.0)
lasttime = time.clock()

# main robot-control loop
while True:

    # S = sense: what do my sensors tell me?
    ALL_SENSORS = robot.GetAllData()
    (x,y,thr,(R,stalled),timestamp,(rx,ry,rthr)) = ALL_SENSORS
    thd = thr*180.0/math.pi   # robot's belief - also x,y,thr
    rthd = rthr*180.0/math.pi # robot's unknown real pose - also rx,ry,rthr
    # in general, using the "real" values is "cheating," since
    # they would not be known to a wandering robot



    # P = plan

    # figure out what state to be in
    # print a message if we are changing state
    if state == GoForward:
        if min(frontThree(R)) < FRONT_THRESH:
            state = TurnRight90
            print 'I am changing to state TurnRight90'
            desiredAngled = thd - 93
            angleWindowd = 3

    elif state == TurnRight90:
            if angleCompare( thd, desiredAngled, angleWindowd ):
                state = GoForward
                print 'I am changing to state GoForward'

    else:
        print 'I\'m lost and so I\'m stopping'
        state = Stop

    # set the velocities appropriate to the state we are in
    v_cmpersec =  velocities[state][0]
    v_degpersec = velocities[state][1]


    # A = act: send velocities to the robot only if they've changed
    if v_cmpersec != v_cm_old or v_degpersec != v_deg_old:
        robot.Go(v_cmpersec, v_degpersec)
        v_cm_old = v_cmpersec
        v_deg_old = v_degpersec


    # visualize: prepare data for the graphics window (robot-depedent)
    #   there is probably a cleverer python way to do this...

    # the next line turns off the orange range sensor bars for now...
    #R = [0,0,0,0,0,0,0]
    msg = 'rB_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f\
                    _%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f' % \
          (x,y,thr,R[0],R[1],R[2],R[3],R[4],R[5],R[6],\
                    rx,ry,rthr,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
                    #rx,ry,rthr,R[0],R[1],R[2],R[3],R[4],R[5],R[6])
    retmsg = vc.exchMessage(msg)

    # check the time
    nowtime = time.clock()
    if nowtime - lasttime > 0.5:
        thr1, dist, thr2 = computeTotalRobotMotion(lastUpdateLoc,(x,y,thr))
        PL = moveParticles( PL, thr1, dist, thr2 )
        PL = updateParticles( PL, robot, ALL_SENSORS )
        vc.sendParticles( PL )
        lastUpdateLoc = (x,y,thr)  # reset where we last updated...
        lasttime = nowtime

    # any keypresses to handle?
    if retmsg == 'pose': # when the robot is reset
        msgToProcess = retmsg
        #print msgToProcess
        msglist = msgToProcess.split('_')
        if msglist[0] == 'pose':
            newx = float(msglist[1])
            newy = float(msglist[2])
            newthr = float(msglist[3])
            robot.SetOdometry(newx, newy, newthr)

    elif retmsg[0:5] == 'char:':
        c = retmsg[5:]
        #print 'I saw a', c
        if c == 'r':
            print 'resetting random particles...'
            # update the particles here...
            PL = randomParticleStart(x,y,thr)
            # these lines will send the message to redraw them...
            vc.sendParticles( PL )
            # remember where the robot was when this reset occurred
            lastUpdateLoc = (x,y,thr)
        elif c == 'm':
            print 'moving particles...'
            thr1, dist, thr2 = computeTotalRobotMotion(lastUpdateLoc,(x,y,thr))
            #print 'thd1 is', math.degrees(thr1)
            #print 'dist is', dist
            #print 'thd2 is', math.degrees(thr2)
            PL = moveParticles( PL, thr1, dist, thr2 )
            # these lines will send the message to redraw them...
            vc.sendParticles( PL )
            # remember where the robot was when this motion occurred
            lastUpdateLoc = (x,y,thr)
        elif c == 'u':
            print 'updating population...'
            PL = updateParticles( PL, robot, ALL_SENSORS )
            print 'len of PL is', len(PL)
            vc.sendParticles( PL )

    # reference vs value exercise...
    # noise exercise = "reasonable model"?
    # hits - how to use them - exceptions?

        # feel free to add responses to your own button presses, if you like!



