# mcl.py
#

# python modules with things we need...
import math                # all the familiar functions
import random              # always fun to have!
import er2sim as er2       # the robot model
import visClient           # to send messages to the Visualizer window

# out map file name
mapfilename= "bugMap.txt"
mapfilename= "mapColors.txt"

# our robot
robot = er2.ER2(PORT=0, mapfile=mapfilename)

# our connection to the robot visualizer
vc = visClient.VisualizerClient()   # using default host and port
vc.exchMessage('loadmap_%s' % mapfilename)

# our connection to the image display
#vc_image = visClient.VisualizerClient(PORT=5009)
#cam_image = visClient.VisualizerClient(PORT=5010)

# the starting robot velocities: translational (cm) and rotational (deg)
v_cm_old = 0.0
v_deg_old = 0.0

# these are constants for accessing the sense range values
BaRi = 0    # the index of the Back Right sensor reading
BaLe = 1    # the index of the Back Left sensor reading
FrLeSi = 2  # the index of the Front Left side-looking sensor reading
FrLe = 3    # the index of the Front Left front-looking sensor reading
Fr   = 4    # the index of the Front sensor reading
FrRi = 5    # the index of the Front Right front-looking sensor reading
FrRiSi = 6  # the index of the Front Right side-looking sensor reading

def frontThree(R):
    return R[3:6]

def angleCompare( thd1, thd2, window ):
    # make sure the angles are between 0.0 and 360.0
    while thd1 > 360.0: thd1 -= 360.0
    while thd2 > 360.0: thd2 -= 360.0
    while thd1 <   0.0: thd1 += 360.0
    while thd2 <   0.0: thd2 += 360.0
    # figure out their relative positions on the unit circle
    smaller = min(thd1,thd2)
    larger  = max(thd1,thd2)
    # figure out the two possible differences
    oneway = larger - smaller
    otherway = (360.0-larger) + smaller
    # is the difference within the window?
    return min(oneway,otherway) <= window

# state constants
GoForward = 0
TurnLeft90 = 1
TurnRight90 = 2
Stop = 8
Goal = 9
MAX_STATE = 10

# initial state
state = GoForward
print 'I am starting in state GoForward'

# constants
FRONT_THRESH = 20

# velocities for each state
velocities = [0]*MAX_STATE
velocities[GoForward] = (42.0, 0.0) # cm/sec, deg/sec
velocities[TurnLeft90] = (0.0, 20.0)
velocities[TurnRight90] = (0.0, -20.0)
velocities[Stop] = (0.0, 0.0)
velocities[Goal] = (0.0, 0.0)


# PL is the particle list
# each particle is a list [x,y,thr,p], where 
# p represents a relative probability (need not sum to 1)
NUM_PARTICLES = 100
def randomParticleUpdate():
    nParts = NUM_PARTICLES
    PL = []
    for i in range(0,nParts):
        # x from -200 to 200
        # y from -100 to 100
        # thr from 0 to 2*pi
        # p is 1.0 (for now)
        PL.append( [ random.randrange(-200,200), \
                     random.randrange(-100,100), \
                     random.randrange(0,360) * math.pi / 180.0, \
                     1.0 ] )
    return PL



# main robot-control loop
while True:

    # S = sense: what do my sensors tell me?
    (x,y,thr,(R,stalled),timestamp,(rx,ry,rthr)) = robot.GetAllData()
    thd = thr*180.0/math.pi   # robot's belief - also x,y,thr
    rthd = rthr*180.0/math.pi # robot's unknown real pose - also rx,ry,rthr
    # in general, using the "real" values is "cheating," since
    # they would not be known to a wandering robot



    # P = plan

    # figure out what state to be in
    # print a message if we are changing state
    if state == GoForward:
	if min(frontThree(R)) < FRONT_THRESH:
            state = TurnRight90
            print 'I am changing to state TurnRight90'
            desiredAngled = thd - 93
            angleWindowd = 3

    elif state == TurnRight90:
            if angleCompare( thd, desiredAngled, angleWindowd ):
                state = GoForward
                print 'I am changing to state GoForward'

    else:
        print 'I\'m lost and so I\'m stopping'
        state = Stop

    # set the velocities appropriate to the state we are in
    v_cmpersec =  velocities[state][0]
    v_degpersec = velocities[state][1]



    # A = act: send velocities to the robot only if they've changed
    if v_cmpersec != v_cm_old or v_degpersec != v_deg_old:
        robot.Go(v_cmpersec, v_degpersec)
        v_cm_old = v_cmpersec
        v_deg_old = v_degpersec


    # visualize: prepare data for the graphics window (robot-depedent)
    #   there is probably a cleverer python way to do this...

    # the next line turns off the orange range sensor bars for now...
    #R = [0,0,0,0,0,0,0]
    msg = 'rB_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f\
                    _%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f' % \
          (x,y,thr,R[0],R[1],R[2],R[3],R[4],R[5],R[6],\
                    rx,ry,rthr,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
                    #rx,ry,rthr,R[0],R[1],R[2],R[3],R[4],R[5],R[6])
    retmsg = vc.exchMessage(msg)

    if retmsg == 'pose': # when the robot is reset
	msgToProcess = retmsg
        #print msgToProcess
        msglist = msgToProcess.split('_')
        if msglist[0] == 'pose':
            newx = float(msglist[1])
            newy = float(msglist[2])
            newthr = float(msglist[3])
            robot.SetOdometry(newx, newy, newthr)

    elif retmsg == 'particleUpdate':
        print 'updating particles...'
        # update the particles here...
        PL = randomParticleUpdate()
        # these lines will send the message to redraw them...
        msg = vc.createParticleMessage(PL)
        vc.exchMessage(msg)



