# mclStart.py
#

# python modules with things we need...
import time                # fun with clocks!
import math                # all the familiar functions
import random              # maybe we'll use it, maybe we won't
import er2sim as er2       # the robot model
import visClient           # to send messages to the Visualizer window
from mclHelp import *      # various angle-processing things

# out map file name
mapfilename= "bugMap.txt"
mapfilename= "mapColors.txt"

# our robot
robot = er2.ER2(PORT=0, mapfile=mapfilename)

# our connection to the robot visualizer
vc = visClient.VisualizerClient()   # using default host and port
vc.exchMessage('loadmap_%s' % mapfilename)

# our connection to the image display
#vc_image = visClient.VisualizerClient(PORT=5009)
#cam_image = visClient.VisualizerClient(PORT=5010)

# the starting robot velocities: translational (cm) and rotational (deg)
v_cm_old = 0.0
v_deg_old = 0.0

# these are constants for accessing the sense range values
BaRi = 0    # the index of the Back Right sensor reading
BaLe = 1    # the index of the Back Left sensor reading
FrLeSi = 2  # the index of the Front Left side-looking sensor reading
FrLe = 3    # the index of the Front Left front-looking sensor reading
Fr   = 4    # the index of the Front sensor reading
FrRi = 5    # the index of the Front Right front-looking sensor reading
FrRiSi = 6  # the index of the Front Right side-looking sensor reading

# state constants
GoForward = 0
TurnLeft90 = 1
TurnRight90 = 2
Stop = 8
Goal = 9
MAX_STATE = 10

# initial state
state = GoForward
print 'I am starting in state GoForward'

# constants
FRONT_THRESH = 20

# velocities for each state
velocities = [0]*MAX_STATE
velocities[GoForward] = (42.0, 0.0) # cm/sec, deg/sec
velocities[TurnLeft90] = (0.0, 20.0)
velocities[TurnRight90] = (0.0, -20.0)
velocities[Stop] = (0.0, 0.0)
velocities[Goal] = (0.0, 0.0)


# PL is the particle list
# each particle is a list [x,y,thr,p], where 
# p represents a relative probability (need not sum to 1)
def randomParticleStart(x=0.0, y=0.0, thr=0.0):
    NUM_PARTICLES = 100
    PL = []
    xrange = 200
    yrange = 200
    thrrange = 0
    for i in range(0,NUM_PARTICLES):
        PL.append( [ random.uniform(x-xrange,x+xrange), \
                     random.uniform(y-yrange,y+yrange), \
                     random.uniform(thr-thrrange,thr+thrrange), \
                     1.0 ] )
    return PL



def moveParticles( PL, thr1, dist, thr2 ):
    for p in PL:
        # create noisy versions
        thr1n = thr1
        distn = random.uniform(41.0,43.0)  # random value from 41.0 to 43.0
        thr2n = thr2
        p[0] += distn   # changes x
        # will have to change p[1], which is y
        # and p[2], which is thr, as well...
    return PL



def resampleParticles( PL ):
    # we want to create a new list of len(PL) particles
    newPL = []   # here's the new list!
    # need the sum of the "probabilities"
    sum = 0.0
    for part in PL: sum += part[3]
    for part in PL: part[3] /= sum
    # now all the particles' probabilities sum to 1.0 
    # but they are NOT all the same
    # you want to get len(PL) particles proportionally from this set...
    # this just gets them randomly (for starters)
    count = 0
    while count < len(PL): 
        i = random.randrange(0,len(PL))
        newPL.append( PL[i][:] )    # the [:] copies the whole particle
        count += 1    # one limitation of python: no count++ or ++count!
    # now return the new population
    return newPL

def updateParticles( PL, robot, ALL_SENSORS ):
    # here is the robot's sensed data
    (x,y,thr,(R,stalled),timestamp,(rx,ry,rthr)) = ALL_SENSORS
    # loop through the particles
    for part in PL:
        # thos obtains each particle's "sensed data" (anticipated)
        RP, numhitsP = robot.GetAllDataFrom(part[0],part[1],part[2]) 
        #
        # RP is the list of range values, just as R is for the robot
        # numhitsP was explained in class...
        part[3] *= 1.0   # bold updating!
    # now resample the particles
    PL = resampleParticles( PL )
    return PL

# update the particles here...
PL = randomParticleStart()
# these lines will send the message to redraw them...
msg = vc.createParticleMessage(PL)
vc.exchMessage(msg)
# remember where the robot was when this reset occurred
lastUpdateLoc = (0.0,0.0,0.0)
lasttime = time.clock()

# main robot-control loop
while True:

    # S = sense: what do my sensors tell me?
    ALL_SENSORS = robot.GetAllData()
    (x,y,thr,(R,stalled),timestamp,(rx,ry,rthr)) = ALL_SENSORS
    thd = thr*180.0/math.pi   # robot's belief - also x,y,thr
    rthd = rthr*180.0/math.pi # robot's unknown real pose - also rx,ry,rthr
    # in general, using the "real" values is "cheating," since
    # they would not be known to a wandering robot



    # P = plan

    # figure out what state to be in
    # print a message if we are changing state
    if state == GoForward:
        if min(frontThree(R)) < FRONT_THRESH:
            state = TurnRight90
            print 'I am changing to state TurnRight90'
            desiredAngled = thd - 93
            angleWindowd = 3

    elif state == TurnRight90:
            if angleCompare( thd, desiredAngled, angleWindowd ):
                state = GoForward
                print 'I am changing to state GoForward'

    else:
        print 'I\'m lost and so I\'m stopping'
        state = Stop

    # set the velocities appropriate to the state we are in
    v_cmpersec =  velocities[state][0]
    v_degpersec = velocities[state][1]


    # A = act: send velocities to the robot only if they've changed
    if v_cmpersec != v_cm_old or v_degpersec != v_deg_old:
        robot.Go(v_cmpersec, v_degpersec)
        v_cm_old = v_cmpersec
        v_deg_old = v_degpersec


    # visualize: prepare data for the graphics window (robot-depedent)
    #   there is probably a cleverer python way to do this...

    # the next line turns off the orange range sensor bars for now...
    #R = [0,0,0,0,0,0,0]
    msg = 'rB_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f\
                    _%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f' % \
          (x,y,thr,R[0],R[1],R[2],R[3],R[4],R[5],R[6],\
                    rx,ry,rthr,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
                    #rx,ry,rthr,R[0],R[1],R[2],R[3],R[4],R[5],R[6])
    retmsg = vc.exchMessage(msg)

    # check the time
    nowtime = time.clock()
    if False and nowtime - lasttime > 0.5:
        thr1, dist, thr2 = computeTotalRobotMotion(lastUpdateLoc,(x,y,thr))
        PL = moveParticles( PL, thr1, dist, thr2 )
        PL = updateParticles( PL, robot, ALL_SENSORS )
        vc.sendParticles( PL )
        lastUpdateLoc = (x,y,thr)  # reset where we last updated...
        lasttime = nowtime

    # any keypresses to handle?
    if retmsg == 'pose': # when the robot is reset
        msgToProcess = retmsg
        #print msgToProcess
        msglist = msgToProcess.split('_')
        if msglist[0] == 'pose':
            newx = float(msglist[1])
            newy = float(msglist[2])
            newthr = float(msglist[3])
            robot.SetOdometry(newx, newy, newthr)

    elif retmsg[0:5] == 'char:':
        c = retmsg[5:]
        #print 'I saw a', c
        if c == 'r':
            print 'resetting random particles...'
            # update the particles here...
            PL = randomParticleStart(x,y,thr)
            # these lines will send the message to redraw them...
            vc.sendParticles( PL )
            # remember where the robot was when this reset occurred
            lastUpdateLoc = (x,y,thr)
        elif c == 'm':
            print 'moving particles...'
            thr1, dist, thr2 = computeTotalRobotMotion(lastUpdateLoc,(x,y,thr))
            print 'thd1 is', math.degrees(thr1)
            print 'dist is', dist
            print 'thd2 is', math.degrees(thr2)
            PL = moveParticles( PL, thr1, dist, thr2 )
            # redraw the particles
            vc.sendParticles( PL )
            # remember where the robot was when this motion occurred
            lastUpdateLoc = (x,y,thr)
        elif c == 'u':
            print 'updating population...'
            PL = updateParticles( PL, robot, ALL_SENSORS )
            vc.sendParticles( PL )

    # reference vs value exercise...
    # noise exercise = "reasonable model"?
    # hits - how to use them - exceptions?

        # feel free to add responses to your own button presses, if you like!



