# spa.py
#
# sense-plan-act loop

# python modules with things we need...
import math     # all the familiar functions
import er2sim as er2       # the robot model
import visClient           # to send messages to the Visualizer window

# out map file name
mapfilename= "bugMap.txt"
mapfilename= "mapColors.txt"

# our robot
robot = er2.ER2(PORT=0, mapfile=mapfilename)

# our connection to the visualizer
vc = visClient.VisualizerClient()   # using default host and port
vc.exchMessage('loadmap_%s' % mapfilename)

# the starting robot velocities: translational (cm) and rotational (deg)
v_cm_old = 0.0
v_deg_old = 0.0

# these are constants for accessing the sense range values
BaRi = 0    # the index of the Back Right sensor reading
BaLe = 1    # the index of the Back Left sensor reading
FrLeSi = 2  # the index of the Front Left side-looking sensor reading
FrLe = 3    # the index of the Front Left front-looking sensor reading
Fr   = 4    # the index of the Front sensor reading
FrRi = 5    # the index of the Front Right front-looking sensor reading
FrRiSi = 6  # the index of the Front Right side-looking sensor reading

def frontThree(R):
    return R[3:6]

def angleCompare( thd1, thd2, window ):
    # make sure the angles are between 0.0 and 360.0
    while thd1 > 360.0: thd1 -= 360.0
    while thd2 > 360.0: thd2 -= 360.0
    while thd1 <   0.0: thd1 += 360.0
    while thd2 <   0.0: thd2 += 360.0
    # figure out their relative positions on the unit circle
    smaller = min(thd1,thd2)
    larger  = max(thd1,thd2)
    # figure out the two possible differences
    oneway = larger - smaller
    otherway = (360.0-larger) + smaller
    # is the difference within the window?
    return min(oneway,otherway) <= window

# state constants
GoForward = 0
TurnLeft90 = 1
TurnRight90 = 2
Stop = 8
Goal = 9
MAX_STATE = 10

# initial state
state = GoForward
state = Stop

# constants
FRONT_THRESH = 20

# velocities for each state
velocities = [0]*MAX_STATE
velocities[GoForward] = (42.0, 0.0) # cm/sec, deg/sec
velocities[TurnLeft90] = (0.0, 20.0)
velocities[TurnRight90] = (0.0, -20.0)
velocities[Stop] = (0.0, 0.0)
velocities[Goal] = (0.0, 0.0)


# main robot-control loop
while True:

    # S = sense: what do my sensors tell me?
    (x,y,thr,(R,stalled),timestamp) = robot.GetAllData()
    thd = thr*180.0/math.pi
    #print 'x,y,thd are', x, y, thd
    #print 'stalled, timestamp, R are', stalled, timestamp, R


    # P = plan

    # figure out what state to be in
    # print a message if we are changing state
    if state == GoForward:
	if min(frontThree(R)) < FRONT_THRESH:
            state = TurnRight90
            print 'I am changing to state TurnRight90'
            desiredAngled = thd - 93
            angleWindowd = 3

    elif state == TurnRight90:
            if angleCompare( thd, desiredAngled, angleWindowd ):
                state = GoForward
                print 'I am changing to state GoForward'

    else:
        # I'm lost - just stop
        state = Stop

    # set the velocities appropriate to the state we are in
    v_cmpersec =  velocities[state][0]
    v_degpersec = velocities[state][1]


    # A = act: send velocities to the robot only if they've changed
    if v_cmpersec != v_cm_old or v_degpersec != v_deg_old:
        robot.Go(v_cmpersec, v_degpersec)
        v_cm_old = v_cmpersec
        v_deg_old = v_degpersec



    # visualize: prepare data for the graphics window (robot-depedent)
    #   there is probably a cleverer python way to do this...
    msg = 'r_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f' %    \
          (x,y,thr,R[0],R[1],R[2],R[3],R[4],R[5],R[6])
    retmsg = vc.exchMessage(msg)
    if retmsg != 'hi 42': 
        print retmsg
        msglist = retmsg.split('_')
        if msglist[0] == 'pose':
            newx = float(msglist[1])
            newy = float(msglist[2])
            newthr = float(msglist[3])
            robot.SetOdometry(newx, newy, newthr)
            # and keep going!

