# spa.py
#
# sense-plan-act loop

# python modules with things we need...
import math     # all the familiar functions
import er2sim as er2       # the robot model
import visClient           # to send messages to the Visualizer window

# our robot
robot = er2.ER2(PORT=0, mapfile=".\\bugMap.txt")

# our connection to the visualizer
vc = visClient.VisualizerClient()   # using default host and port
vc.sendMessage('loadmap_bugMap.txt')

# the starting robot velocities: translational (cm) and rotational (deg)
v_cm_old = 0.0
v_deg_old = 0.0

# these are constants for accessing the sense range values
BaRi = 0    # the index of the Back Right sensor reading
BaLe = 1    # the index of the Back Left sensor reading
FrLeSi = 2  # the index of the Front Left side-looking sensor reading
FrLe = 3    # the index of the Front Left front-looking sensor reading
Fr   = 4    # the index of the Front sensor reading
FrRi = 5    # the index of the Front Right front-looking sensor reading
FrRiSi = 6  # the index of the Front Right side-looking sensor reading

v_cm_last = 35.0


# wander "forever"
while True:

    # S = sense: what do my sensors tell me?
    (x,y,thr,(R,stalled),timestamp) = robot.GetAllData()
    #print 'x,y,thd are', x, y, thr*180.0/math.pi
    #print '   stalled, timestamp, R are', stalled, timestamp, R


    # P = plan

    # reverse if there is a collision (the robot has "stalled")
    if stalled:
        v_cm_last = -v_cm_last

    v_cmpersec =  v_cm_last
    v_degpersec = 0.0


    # A = act: send velocities to the robot only if they've changed
    if v_cmpersec != v_cm_old or v_degpersec != v_deg_old:
        robot.Go(v_cmpersec, v_degpersec)
        v_cm_old = v_cmpersec
        v_deg_old = v_degpersec



    # visualize: prepare data for the graphics window (robot-depedent)
    #   there is probably a cleverer python way to do this...
    msg = 'r_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f_%.2f' %    \
          (x,y,thr,R[0],R[1],R[2],R[3],R[4],R[5],R[6])
    vc.sendMessage(msg)

